/************************************************************************
* Copyright (C) 2000 Thomas Schulz					*
*									*/
	static char rcsid[] =
	"$Id: comio.c,v 1.7 2001/01/17 23:43:51 tsch Rel $";  	/*
*									*
* IC35 serial communication
*									*
*************************************************************************
*									*
*	??? is "fixme" mark: sections of code needing fixes		*
*									*
*		simulated communication					*
*	com_siminit	initialize communications with simulation file	*
*	com_simexit	local: leave simulated communication		*
*	com_simul	local: report if simulation active		*
*	com_simrecv	local: simulate receive from simulation file	*
*		real communication
*	com_setsigs	local: set RS232 output signals			*
*	com_settimeout	set receive timeout, return previous		*
*	com_init	initialize serial communication device		*
*	com_sendw	send datablock to comm.device with waiting	*
*	com_send	send datablock to comm.device			*
*	com_recv	receive datablock from comm.device		*
*	com_exit	close serial communication device		*
*									*
************************************************************************/

#include <stdio.h>	/* FILE*, fopen(), ..	*/
#include <string.h>	/* strncmp(), ..	*/
#include <unistd.h>	/* read(), write(), ..	*/
#include <termios.h>	/* tcgetattr(), ..	*/
#include <fcntl.h>	/* F_GETFL, ..		*/
#include <sys/ioctl.h>	/* ioctl(), ..		*/
#include <sys/types.h>	/* size_t, ..		*/
#include <sys/time.h>	/* struct timeval	*/
#include <signal.h>	/* sigaction(), ..	*/

#include "util.h"	/* lprintf(), ..	*/
#include "comio.h"
NOTUSED(rcsid)


/* ==================================== */
/*	simulated communication		*/
/* ==================================== */
/*
*	uses post-processed log of real communication with IC35:
*	  WR nn   xx xx xx ...
*	  RD nn   xx xx xx ...
*	and "receives" data using the "RD nn" lines.
*/
static char *	simfname = NULL;
static FILE *	simfp    = NULL;

void
com_siminit( char * s_fname )			/* init simulated comm	*/
{
    if ( s_fname && *s_fname )
	simfp = fopen( simfname = s_fname, "r" );
}
static void
com_simexit( void )				/* leave simulated comm	*/
{
    if ( simfp ) {
	fclose( simfp );
	simfp = NULL;
	simfname = NULL;
    }
}
static bool
com_simul( void )				/* report if simul active  */
{
    return (bool)( simfp != NULL );
}
static int
com_simrecv( uchar * buff, size_t blen )	/* receive from simul.file */
{
    static int	simrlen = 0;
    static int	simridx = 0;
    uchar *	bptr;
    int		chr, n, rbyte;
    char	xdir[8];

			/* forward to next "RD nn" line	*/
    if ( simridx >= simrlen ) {
	do {
	    while ( (chr = fgetc( simfp )) != '\n' )
		if ( chr < 0 )
		    return ERR;
	    n = fscanf( simfp, "%s %d", xdir, &simrlen );
	} while ( n > 0 && strncmp( xdir, "RD", 2 ) != 0 );
	if ( n <= 0 )
	    return ERR;
	simridx = 0;
    }
			/* read bytes from "RD nn" line	*/
    for ( bptr = buff; bptr < buff+blen; ++bptr ) {
	if ( simridx >= simrlen )
	    break;
	if ( fscanf( simfp, "%x", &rbyte ) < 0 )
	    break;
	*bptr = (uchar)rbyte;
	++simridx;
    }
    return bptr - buff;
}


/* ==================================== */
/*	real communication		*/
/* ==================================== */

static int	com_fd = -1;
static int	com_tmo = 500;		/* timeout 500 ms	*/

/* local: set RS232 output signals
*  -------------------------------
*/
static void
com_setsigs( int sigs )
{
    int 	flags;

    if ( ioctl( com_fd, TIOCMGET, &flags ) == 0 ) {
	flags &= ~(TIOCM_DTR|TIOCM_RTS);
	flags |=  (TIOCM_DTR|TIOCM_RTS) & sigs;
	ioctl( com_fd, TIOCMSET, &flags );
	lprintf( L_NOISE, "com_setsigs(%08X) DTR %s RTS %s",
			    sigs, sigs & TIOCM_DTR ? "ON " : "off",
				  sigs & TIOCM_RTS ? "ON " : "off" );
    }
}

/* local: log state of RS232 signals
*  ---------------------------------
*/
static void
_com_sigchg( void )
{
    static int	oflags = -1;
    static struct {
	char *	  name;
	int	  sig;
    }		sigtab[] = {
		    { "CTS", TIOCM_CTS	},
		    { "DCD", TIOCM_CAR	},
		    { "DSR", TIOCM_DSR	},
		    { "RI",  TIOCM_RNG	},
		    {  NULL,	0	}
		}, *psig;
    int 	flags;
    char	sigtext[48];

    if ( com_fd >= 0
      && ioctl( com_fd, TIOCMGET, &flags ) == 0
      && ( ((oflags ^ flags) & (TIOCM_CTS|TIOCM_CAR|TIOCM_DSR|TIOCM_RNG)) != 0
	  || oflags == -1 ) ) {
	strcpy( sigtext, "" );
	for ( psig = sigtab; psig->name; ++psig )
	    sprintf( sigtext+strlen(sigtext), " %s %s%s",
		    psig->name,
		    flags & psig->sig ? "ON" : "off",
		    ((oflags ^ flags) & psig->sig) && oflags != -1 ? "*" : "" );
	lprintf( L_NOISE, "com signals:%s", sigtext );
	oflags = flags;
    }
}

/* set,get receive timeout
*  -----------------------
*	return previous timeout. zero or negative timeout argument 'msec'
*	will not be set and can be used to enquire current timeout.
*/
int
com_settimeout( int msec )
{
    int 	old_tmo = com_tmo;

    old_tmo = com_tmo;
    if ( msec > 0 )
	com_tmo = msec;
    if ( old_tmo != com_tmo )
	lprintf( L_NOISE, "com_settimeout %d -> %d", old_tmo, com_tmo );
    return old_tmo;
}

/* initialize comm.device
*  ----------------------
*	open in O_NONBLOCK mode to avoid hanging on inactice DCD signal,
*	finally set back to blocking mode.
*	set line parameters to 115200,N,8,2 and raw mode, 2 stopbits are
*	needed to avoid IC35 receive errors with MMCard operations.
*	raise DTR and RTS signals.
*/
int
com_init( char * devname )
{
    int 		i, flags;
    struct termios	tio;

    if ( com_simul() ) {
	com_fd = -1;
	return OK;
    }
	/* open serial communication device	*/
    if ( (com_fd = open( devname,  O_RDWR|O_NONBLOCK )) == -1 ) {
	return ERR;
    }
    if ( !isatty( com_fd ) ) {
	close( com_fd );
	com_fd = -1;
	return ERR;
    }
	/* set linepars 11500,N,8,2 and raw mode */
	/* HUPCL to clear DTR,RTS on last close  */
    tcgetattr( com_fd, &tio );
    tio.c_oflag = 0;
    tio.c_iflag = IGNBRK | IGNPAR;
    tio.c_cflag = CREAD | CLOCAL | CS8 | HUPCL | CSTOPB;
    cfsetispeed( &tio, B115200 );
    cfsetospeed( &tio, B115200 );
    tio.c_lflag = NOFLSH;
    cfmakeraw( &tio );
    for ( i = 0; i <= NCCS; ++i ) tio.c_cc[i] = 0;
    tio.c_cc[VMIN] = 1;
    tio.c_cc[VTIME] = 0;
    tcsetattr( com_fd, TCSANOW, &tio );
	/* set line signals DTR on, RTS off	*/
    com_setsigs( TIOCM_DTR|TIOCM_RTS );
	/* back to blocking mode for timeout	*/
    if ( (flags = fcntl( com_fd, F_GETFL, 0 )) != -1 ) {
	flags &= ~O_NONBLOCK;
	fcntl( com_fd, F_SETFL, flags );
    }
    _com_sigchg();

    return OK;
}

/* send datablock with waiting
*  ---------------------------
*	IC35 gets receive receive errors if datablocks of more than 16 bytes
*	are sent at full speed of 115200 baud, using 2 stopbits increases
*	this size limit to 29 bytes.
*	the problem occurred only with IC35 MMCard operations, import and
*	synchronize PIM-data don't suffer this limitation.
*	as a workaround do delay some milliseconds after sent every 29 bytes.
*	the minimum delay possible with Linux-2.x is ca. 10 msec by using
*	an "itimer" (due to Linux timer interrupt frequency of 100 HZ).
*	other methods like usleep(), nanosleep(), select() yielded delays
*	of ca. 20..30 msec and were therefore not chosen.
*	if the "itimer" SIGALRM arrives too early (measured up to 7 msec),
*	the delay is continued. elapsed delay time up to 2 msec less than
*	the requested delay time was tested OK.
*	see getitimer(2), sigaction(2) (and also nanosleep(2), usleep(3)
*	select for compare).
*	the maximum throughput measured with write file to IC35 MMCard
*	was ca. 1350 b/s (with usleep() delay it was only ca. 600 b/s).
*/
static struct sigaction _oldsigact;
static struct itimerval _olditimer;
static int		_ntimerint;
static unsigned		_timerusec;

static void
_timer_hdlr( int sig )
{
    ++_ntimerint;
}
static void
_start_timer( unsigned usec )
{
    struct sigaction	sigact;
    struct itimerval	itimer;

    _ntimerint = 0;
    _timerusec = usec;
    memset( &sigact, 0, sizeof(sigact) );
    sigact.sa_handler = _timer_hdlr;
    sigact.sa_flags   = SA_RESTART;
    sigaction( SIGALRM, &sigact, &_oldsigact );

    itimer.it_interval.tv_sec  = usec / 1000000;
    itimer.it_interval.tv_usec = usec % 1000000;
    itimer.it_value = itimer.it_interval;
    setitimer( ITIMER_REAL, &itimer, &_olditimer );
}
static void
_wait_timer( void )
{
    struct timeval	tend, tnow;

    gettimeofday( &tend, NULL );
    tend.tv_sec  += (_timerusec - 2000) / 1000000;	/* early by 2 msec */
    tend.tv_usec += (_timerusec - 2000) % 1000000;	/*  was tested OK  */
    if ( tend.tv_usec >= 1000000 ) {
	tend.tv_usec  -= 1000000;
	tend.tv_sec   += 1;
    }
    do {
	pause();				/* wait for itimer SIGALRM  */
	gettimeofday( &tnow, NULL );
    } while ( tnow.tv_sec  <= tend.tv_sec	/* until delay time elapsed */
	   && tnow.tv_usec <= tend.tv_usec );
}
static void
_stop_timer( void )
{
    setitimer( ITIMER_REAL, &_olditimer, NULL );
    sigaction( SIGALRM, &_oldsigact, NULL );
    lprintf( L_DEBUG, "_stop_timer: got %d itimer signals", _ntimerint );
}

int
com_sendw( uchar * data, size_t dlen )	/* send data to comm.dev and wait */
{
    uchar *	dptr;
    int 	slen;

    if ( !com_simul() ) {
	if ( com_fd < 0 )
	    return ERR;
	_start_timer( 10000 );
	for ( dptr = data; dptr < data+dlen; ++dptr ) {
	    if ( (dptr - data) % 29 == 0 ) {
		_wait_timer();
		lprintf( L_DEBUG, "com_send: write(com,data+%04X,)",
							dptr - data );
	    }
	    if ( write( com_fd, dptr, 1 ) != 1 )
		break;
	}
	_stop_timer();
	slen = dptr - data;
    } else {
	slen = dlen;
    }
    ldump( L_NOISE, data, dlen,
		    "com_send(%p,%u) = %d", data, dlen, slen );
    return slen;
}

/* send datablock to comm.device
*  -----------------------------
*/
int
com_send( uchar * data, size_t dlen )
{
    int 	slen;

    if ( !com_simul() ) {
	if ( com_fd < 0 )
	    return ERR;
	slen = write( com_fd, data, dlen );
    } else {
	slen = dlen;
    }
    ldump( L_NOISE, data, dlen,
		    "com_send(%p,%u) = %d", data, dlen, slen );
    return slen;
}

/* receive datablock from comm.device
*  ----------------------------------
*	use select() to wait for data, accumulate data with read()
*	until buffer 'buff' gets full or receive timeout occurs.
*/
int
com_recv( uchar * buff, size_t blen )
{
    fd_set		rfds;
    struct timeval	tmo;
    uchar		*rptr, *rend;
    int 		rlen;

    if ( !com_simul() ) {
	if ( com_fd < 0 )
	    return ERR;
	rend = (rptr = buff) + blen;
	do {
	    FD_ZERO( &rfds );
	    FD_SET( com_fd, &rfds );
	    tmo.tv_sec  =  com_tmo / 1000;
	    tmo.tv_usec = (com_tmo % 1000) * 1000;
	    if ( select( com_fd+1, &rfds, NULL, NULL, &tmo ) > 0
	      && FD_ISSET( com_fd, &rfds ) ) {
		_com_sigchg();
		rlen = read( com_fd, rptr, rend - rptr );
		lprintf( L_DEBUG, "com_recv: read(com,rptr,%d) = %d",
							rend-rptr, rlen );
		if ( rlen > 0 )
		    rptr += rlen;
	    } else {
		rlen = 0;
	    }
	} while ( rlen > 0 && rptr < rend );
	rlen = rptr - buff;
    } else {
	rlen = com_simrecv( buff, blen );
    }
    if ( rlen > 0 )
	ldump( L_NOISE, buff, rlen,
			  "com_recv(%p,%u) = %d", buff, blen, rlen );
    else
	lprintf( L_NOISE, "com_recv(%p,%u) = %d", buff, blen, rlen );

    return rlen;
}

/* close comm.device
*  -----------------
*	leave simulated communication if active.
*	clear RS232 signals DTR and RTS and close comm.device.
*/
void
com_exit( void )
{
    com_simexit();
    if ( com_fd < 0 )
	return;
    com_setsigs( 0 );				/* DTR off, RTS off */
    _com_sigchg();
    close( com_fd );
    com_fd = -1;
}
